@@ -9,37 +9,34 @@
 #   add meta-data in tuner's returns
 #   add "tradingview backtest"
 #   multi-timeframe backtest
+#   telegram bot
+#   https://smart-lab.ru/company/www-marketstat-ru/blog/502764.php
 
-from copy import copy
+from copy import deepcopy
 from datetime import datetime
 from re import fullmatch
 from threading import Thread
-from time import ctime
-from time import sleep
-from time import time
-from typing import Any
-from typing import Dict
-from typing import Iterable
-from typing import List
-from typing import Tuple
-from typing import Union
+from time import ctime, sleep, time
+from typing import Any, Union, List, Iterable, Tuple, Dict
 from warnings import warn
 
+import numpy as np
+import pandas as pd
 import ta.momentum
 import ta.trend
 import ta.volatility
-import numpy as np
-import pandas as pd
 
+from . import indicators
 from . import utils
 from .brokers import TradingClient
-from .plots import QuickTradeGraph
+from .plots import TraderGraph
+from . import strategy
 
 
 class Trader(object):
     _profit_calculate_coef: Union[float, int]
     returns: utils.PREDICT_TYPE_LIST = []
-    df: pd.DataFrame
+    _df: pd.DataFrame
     ticker: str
     interval: str
     _prev_predict: str = 'Exit'
@@ -49,68 +46,80 @@ class Trader(object):
     trades: int = 0
     profits: int = 0
     losses: int = 0
-    _stop_losses: List[float]
-    _take_profits: List[float]
-    _credit_leverages: List[Union[float, int]]
-    deposit_history: List[Union[float, int]]
+    stop_losses: List[float] = []
+    take_profits: List[float] = []
+    credit_leverages: List[Union[float, int]] = []
+    deposit_history: List[Union[float, int]] = []
     year_profit: float
     _info: str
     backtest_out: pd.DataFrame
-    _open_lot_prices: List[float]
+    open_lot_prices: List[float] = []
     client: TradingClient
     __last_stop_loss: float
     __last_take_profit: float
-    net_returns: pd.Series
     _sec_interval: int
     supports: Dict[int, float]
     resistances: Dict[int, float]
-    fig: QuickTradeGraph
+    fig: TraderGraph
     _multi_converted_: bool = False
     _entry_start_trade: bool
+    average_growth: Union[np.ndarray, List]
+    _converted: utils.CONVERTED_TYPE_LIST
+    mean_deviation: float
+    sharpe_ratio: float
+    sortino_ratio: float
+    max_drawdown: float
+    calmar_ratio: float
+    net_returns: pd.Series
+    profit_deviation_ratio: float
+    _registered_strategy: str
 
-    @property
-    def _converted(self) -> utils.CONVERTED_TYPE_LIST:
-        return utils.convert(self.returns)
+    def returns_update(self):
+        self._converted = utils.convert(self.returns)
 
-    @property
-    def mean_deviation(self) -> float:
-        return utils.mean_deviation(pd.Series(self.deposit_history), self.average_growth) * 100
+    def deposit_history_update(self):
+        self.average_growth = utils.get_exponential_growth(self.deposit_history)
+        self.mean_deviation = utils.mean_deviation(pd.Series(self.deposit_history), self.average_growth) * 100
 
-    @property
-    def cumulative_returns(self) -> pd.Series:
-        return pd.Series(self.deposit_history) / self.deposit_history[0]
-
-    @property
-    def sharpe_ratio(self) -> float:
+        # Sharpe ratio
         returns = utils.get_multipliers(pd.Series(self.deposit_history)) - 1
         mean_ = returns.mean() * self._profit_calculate_coef
         sigma = returns.std() * np.sqrt(self._profit_calculate_coef)
-        return mean_ / sigma
+        self.sharpe_ratio = mean_ / sigma
 
-    @property
-    def sortino_ratio(self) -> float:
-        returns = utils.get_multipliers(pd.Series(self.deposit_history)) - 1
-        mean_ = returns.mean() * self._profit_calculate_coef
+        # Sortino ratio
         sigma = returns[returns < 0].std() * np.sqrt(self._profit_calculate_coef)
-        return mean_ / sigma
+        self.sortino_ratio = mean_ / sigma
+
+        # max drawdown
+        max_returns = np.fmax.accumulate(self.deposit_history)
+        res = self.deposit_history / max_returns - 1
+        self.max_drawdown = abs(min(res) * 100)
+
+        # average year profit
+        self.year_profit = utils.year_profit(self.average_growth, self._profit_calculate_coef)
+
+        if self.trades != 0:
+            self.winrate = (self.profits / self.trades) * 100
+        else:
+            self.winrate = 0
+            utils.logger.critical('0 trades in %s', self)
+
+        self.calmar_ratio = self.year_profit / self.max_drawdown
+
+        # net returns
+        self.net_returns = pd.Series(self.deposit_history).diff()
+        self.net_returns[0] = 0.0
+
+        self.profit_deviation_ratio = self.year_profit / self.mean_deviation
 
     @property
-    def max_drawdown(self):
-        max_returns = np.fmax.accumulate(self.cumulative_returns)
-        res = self.cumulative_returns / max_returns - 1
-        return abs(min(res) * 100)
+    def df(self) -> pd.DataFrame:
+        return self._df
 
-    @property
-    def calmar_ratio(self):
-        return self.year_profit / self.max_drawdown
-
-    @property
-    def average_growth(self) -> np.ndarray:
-        return utils.get_exponential_growth(self.deposit_history)
-
-    @property
-    def profit_deviation_ratio(self):
-        return self.year_profit / self.mean_deviation
+    @df.setter
+    def df(self, frame: pd.DataFrame):
+        self._df = frame
 
     @property
     def _info(self):
@@ -204,9 +213,9 @@ class Trader(object):
         utils.logger.debug('add stop-loss: %f pips, take-profit: %s pips', add_stop_loss, add_take_profit)
         stop_losses = []
         take_profits = []
-        for stop_loss_price, take_profit_price, price, sig in zip(self._stop_losses,
-                                                                  self._take_profits,
-                                                                  self._open_lot_prices,
+        for stop_loss_price, take_profit_price, price, sig in zip(self.stop_losses,
+                                                                  self.take_profits,
+                                                                  self.open_lot_prices,
                                                                   self.returns):
             add_sl = (price / 10_000) * add_stop_loss
             add_tp = (price / 10_000) * add_take_profit
@@ -221,17 +230,17 @@ class Trader(object):
                 stop_losses.append(stop_loss_price)
                 take_profits.append(take_profit_price)
 
-        self._stop_losses = stop_losses
-        self._take_profits = take_profits
-        return self._stop_losses, self._take_profits
+        self.stop_losses = stop_losses
+        self.take_profits = take_profits
+        return self.stop_losses, self.take_profits
 
     def multi_trades(self):
-        self.returns, self._credit_leverages = utils.make_multi_trade_returns(self.returns)
+        self.returns, self.credit_leverages = utils.make_multi_trade_returns(self.returns)
         self._multi_converted_ = True
 
     def get_heikin_ashi(self, df: pd.DataFrame = pd.DataFrame()) -> pd.DataFrame:
         """
-        :param df: dataframe, standard: self.df
+        :param df: dataframe, default: self.df
         :return: heikin ashi
         """
         if 'Close' not in df.columns:
@@ -288,8 +297,9 @@ class Trader(object):
                 flag = utils.EXIT
             returns.append(flag)
         self.returns = returns
+        self.returns_update()
         if swap_stop_take:
-            self._stop_losses, self._take_profits = self._take_profits, self._stop_losses
+            self.stop_losses, self.take_profits = self.take_profits, self.stop_losses
         return self.returns
 
     @utils.assert_logger
@@ -320,6 +330,9 @@ class Trader(object):
         assert isinstance(print_out, bool), 'print_out must be of type <bool>'
         assert isinstance(show, bool), 'show must be of type <bool>'
 
+        self.returns_update()
+        pass_math: bool = False
+        data_column: pd.Series = self.df['Close']
         exit_take_stop: bool
         no_order: bool
         stop_loss: float
@@ -332,7 +345,6 @@ class Trader(object):
         credit_lev: Union[float, int]
 
         start_bet: Union[float, int] = bet
-        data_column: pd.Series = self.df['Close']
         data_high: pd.Series = self.df['High']
         data_low: pd.Series = self.df['Low']
         self.deposit_history = [deposit]
@@ -355,7 +367,6 @@ class Trader(object):
         low: float
         next_h: float
         next_l: float
-        pass_math: bool = False
         normal: bool
         for e, (sig,
                 stop_loss,
@@ -366,10 +377,10 @@ class Trader(object):
                 low,
                 next_h,
                 next_l) in enumerate(zip(self.returns[:-1],
-                                         self._stop_losses[:-1],
-                                         self._take_profits[:-1],
+                                         self.stop_losses[:-1],
+                                         self.take_profits[:-1],
                                          self._converted[:-1],
-                                         self._credit_leverages[:-1],
+                                         self.credit_leverages[:-1],
                                          data_high[:-1],
                                          data_low[:-1],
                                          data_high[1:],
@@ -405,29 +416,22 @@ class Trader(object):
                 ignore_breakout = True
 
                 if sig != utils.EXIT and not min(stop_loss, take_profit) <= open_price <= max(stop_loss, take_profit) and e > 0:
-                    warn('The deal was opened out of range!')
-                    utils.logger.error('(%s) The deal was opened out of range! (candle=%d)', self, e)
-                    self.winrate = 0.0
-                    self.year_profit = 0.0
-                    self.losses = 0
-                    self.profits = 0
-                    self.trades = 0
                     pass_math = True
                     break
 
             if sig != utils.EXIT:
                 next_not_breakout = min(stop_loss, take_profit) < next_l <= next_h < max(stop_loss, take_profit)
 
-                stop_loss = self._stop_losses[e - 1]
-                take_profit = self._take_profits[e - 1]
+                stop_loss = self.stop_losses[e - 1]
+                take_profit = self.take_profits[e - 1]
                 # be careful with e=0
                 # haha))) no)
                 now_not_breakout = min(stop_loss, take_profit) < low <= high < max(stop_loss, take_profit)
 
                 normal = ignore_breakout or (now_not_breakout and next_not_breakout)
 
-                if credit_lev != self._credit_leverages[e - 1] and not ignore_breakout:
-                    deposit -= bet * (commission / 100) * abs(self._credit_leverages[e - 1] - credit_lev)
+                if credit_lev != self.credit_leverages[e - 1] and not ignore_breakout:
+                    deposit -= bet * (commission / 100) * abs(self.credit_leverages[e - 1] - credit_lev)
                     # Commission when changing the leverage.
                     if bet > deposit:
                         bet = deposit
@@ -450,8 +454,8 @@ class Trader(object):
                     exit_take_stop = True
 
                     if (not now_not_breakout) and not ignore_breakout:
-                        stop_loss = self._stop_losses[e - 1]
-                        take_profit = self._take_profits[e - 1]
+                        stop_loss = self.stop_losses[e - 1]
+                        take_profit = self.take_profits[e - 1]
                         diff = utils.get_diff(price=data_column[e],
                                               low=low,
                                               high=high,
@@ -460,8 +464,8 @@ class Trader(object):
                                               signal=sig)
 
                     elif not next_not_breakout:
-                        stop_loss = self._stop_losses[e]
-                        take_profit = self._take_profits[e]
+                        stop_loss = self.stop_losses[e]
+                        take_profit = self.take_profits[e]
                         diff = utils.get_diff(price=data_column[e],
                                               low=next_l,
                                               high=next_h,
@@ -483,29 +487,28 @@ class Trader(object):
                 prev_sig = sig
             ignore_breakout = False
 
-        self.net_returns = pd.Series(self.deposit_history).diff()
-        self.net_returns[0] = 0
-        if not pass_math:
-            self.year_profit = utils.profit_factor(self.average_growth) ** (self._profit_calculate_coef - 1)
-            #  Compound interest. View https://www.investopedia.com/terms/c/compoundinterest.asp
-            self.year_profit -= 1  # The initial deposit does not count as profit
-            self.year_profit *= 100  # Percentage
-            if self.trades != 0:
-                self.winrate = (self.profits / self.trades) * 100
-            else:
-                self.winrate = 0
-                utils.logger.critical('0 trades in %s', self)
-        utils.logger.info('(%s) trader info: %s', self, self._info)
+        self.deposit_history_update()
+
+        if pass_math:
+            warn('The deal was opened out of range!')
+            utils.logger.error('(%s) The deal was opened out of range! (candle=%d)', self, e)
+            self.winrate = 0.0
+            self.year_profit = 0.0
+            self.losses = 0
+            self.profits = 0
+            self.trades = 0
+
         if print_out:
             print(self._info)
+        utils.logger.info('(%s) trader info: %s', self, self._info)
         self.backtest_out = pd.DataFrame(
-            (self.deposit_history, self._stop_losses, self._take_profits, self.returns,
-             self._open_lot_prices, data_column, self.average_growth, self.net_returns, self.cumulative_returns),
+            (self.deposit_history, self.stop_losses, self.take_profits, self.returns,
+             self.open_lot_prices, data_column, self.average_growth, self.net_returns),
             index=[
                 'deposit', 'stop loss', 'take profit',
                 'predictions', 'open trade', 'Close',
                 "average growth deposit data",
-                "returns", 'cumulative returns'
+                "returns"
             ]).T
         if plot:
             self.fig.plot_candlestick()
@@ -530,7 +533,8 @@ class Trader(object):
                        commission: Union[float, int] = 0.0,
                        plot: bool = True,
                        print_out: bool = True,
-                       show: bool = True) -> pd.DataFrame:
+                       show: bool = True,
+                       _dataframes=None) -> pd.DataFrame:
         for el in test_config.keys():
             assert isinstance(el, str), 'tickers must be of type <Iterable[str]>'
             assert fullmatch(utils.TICKER_PATTERN, el), f'all tickers must match the pattern <{utils.TICKER_PATTERN}>'
@@ -552,7 +556,6 @@ class Trader(object):
         assert isinstance(show, bool), 'show must be of type <bool>'
 
         winrates: List[float] = []
-        percentage_profits: List[float] = []
         losses: List[int] = []
         trades: List[int] = []
         profits: List[int] = []
@@ -563,11 +566,14 @@ class Trader(object):
         for ticker, strat_l in test_config.items():
             for strat in strat_l:
                 for strategy_kwargs in strat.items():
-                    df = self.client.get_data_historical(ticker=ticker, limit=limit, interval=self.interval)
+                    if _dataframes is None:
+                        df = self.client.get_data_historical(ticker=ticker, limit=limit, interval=self.interval)
+                    else:
+                        df = _dataframes[ticker]
                     new_trader = self._get_this_instance(interval=self.interval, df=df, ticker=ticker)
-                    new_trader.set_client(your_client=self.client)
+                    new_trader.set_client(client=self.client)
                     try:
-                        new_trader.connect_graph(copy(self.fig))
+                        new_trader.connect_graph(deepcopy(self.fig))
                     except Exception as e:
                         if isinstance(e, KeyboardInterrupt):
                             raise e
@@ -580,16 +586,15 @@ class Trader(object):
                                         print_out=False,
                                         show=False)
                     winrates.append(new_trader.winrate)
-                    percentage_profits.append(new_trader.year_profit)
                     losses.append(new_trader.losses)
                     trades.append(new_trader.trades)
                     profits.append(new_trader.profits)
                     depos.append(pd.Series(new_trader.deposit_history))
                     lens_dep.append(len(new_trader.deposit_history))
+        self._registered_strategy = str(test_config)
         self.losses = sum(losses)
-        self.trades = sum(trades)
+        self.trades = sum(trades)  # all time
         self.profits = sum(profits)
-        self.year_profit = float(np.mean(percentage_profits))
         self.winrate = float(np.mean(winrates))
 
         for enum, elem in enumerate(depos):
@@ -598,15 +603,13 @@ class Trader(object):
         multipliers: pd.Series = sum(depos) / len(depos)
         multipliers[0] = deposit
         self.deposit_history = list(np.cumprod(multipliers.values))
-        self.net_returns = pd.Series(self.deposit_history).diff()
-        self.net_returns[0] = 0
+        self.deposit_history_update()
         self.backtest_out = pd.DataFrame(
-            (self.deposit_history, self.average_growth, self.net_returns, self.cumulative_returns),
+            (self.deposit_history, self.average_growth, self.net_returns),
             index=[
                 'deposit',
                 "average growth deposit data",
                 "returns",
-                'cumulative returns'
             ]).T
 
         utils.logger.info('(%s) trader multi info: %s', self, self._info)
@@ -621,9 +624,9 @@ class Trader(object):
 
     @utils.assert_logger
     def connect_graph(self,
-                      graph: QuickTradeGraph):
+                      graph: TraderGraph):
         """
-        connect QuickTradeGraph
+        connect TraderGraph
         """
 
         self.fig = graph
@@ -766,10 +769,10 @@ class Trader(object):
         predict: str = utils.convert_signal_str(self.returns[-1])
 
         # trading
-        self.__last_stop_loss = self._stop_losses[-1]
-        self.__last_take_profit = self._take_profits[-1]
+        self.__last_stop_loss = self.stop_losses[-1]
+        self.__last_take_profit = self.take_profits[-1]
 
-        conv_cred_lev = utils.convert(self._credit_leverages)
+        conv_cred_lev = utils.convert(self.credit_leverages)
 
         if self._entry_start_trade:
             open_new_order = predict != self._prev_predict or not np.isnan(conv_cred_lev[-1])
@@ -799,7 +802,7 @@ class Trader(object):
 
                         self.client.order_create(predict,
                                                  self.ticker,
-                                                 bet * self._credit_leverages[-1])  # entry new position
+                                                 bet * self.credit_leverages[-1])  # entry new position
         utils.logger.debug("(%s) returning prediction", self)
         return {
             'predict': predict,
@@ -807,7 +810,7 @@ class Trader(object):
             'stop loss': self.__last_stop_loss,
             'take profit': self.__last_take_profit,
             'close price': close[-1],
-            'credit leverage': self._credit_leverages[-1]
+            'credit leverage': self.credit_leverages[-1]
         }
 
     @utils.assert_logger
@@ -950,8 +953,8 @@ class Trader(object):
         def start_trading(pair, strat):
             trader = MultiRealTimeTrader(ticker=pair,
                                          interval=self.interval)
-            trader.connect_graph(graph=copy(self.fig))
-            trader.set_client(copy(client))
+            trader.connect_graph(graph=deepcopy(self.fig))
+            trader.set_client(deepcopy(client))
 
             items = tuple(strat.items())
             item = items[0]
@@ -987,13 +990,13 @@ class Trader(object):
         utils.logger.info('(%s) log returns', self)
 
     @utils.assert_logger
-    def set_client(self, your_client: TradingClient):
+    def set_client(self, client: TradingClient):
         """
-        :param your_client: trading client
+        :param client: trading client
         """
-        assert isinstance(your_client, TradingClient), 'your_client must be of type <TradingClient>'
+        assert isinstance(client, TradingClient), 'client must be of type <TradingClient>'
 
-        self.client = your_client
+        self.client = client
         utils.logger.info('(%s) set client', self)
 
     @utils.assert_logger
@@ -1028,15 +1031,16 @@ class Trader(object):
         assert isinstance(set_stop, bool), 'set_stop must be of type <bool>'
         assert isinstance(set_take, bool), 'set_stop must be of type <bool>'
 
+        self.returns_update()
         self._take_profit = take_profit
         self._stop_loss = stop_loss
         take_flag: float = np.inf
         stop_flag: float = np.inf
-        self._open_lot_prices = []
+        self.open_lot_prices = []
         if set_stop:
-            self._stop_losses = []
+            self.stop_losses = []
         if set_take:
-            self._take_profits = []
+            self.take_profits = []
         closes: np.ndarray = self.df['Close'].values
         sig: utils.PREDICT_TYPE
         close: float
@@ -1055,15 +1059,15 @@ class Trader(object):
                 else:
                     take_flag = stop_flag = self._open_price
 
-            self._open_lot_prices.append(self._open_price)
+            self.open_lot_prices.append(self._open_price)
             if set_take:
-                self._take_profits.append(take_flag)
+                self.take_profits.append(take_flag)
             elif sig == utils.EXIT:
-                self._take_profits[e] = take_flag
+                self.take_profits[e] = take_flag
             if set_stop:
-                self._stop_losses.append(stop_flag)
+                self.stop_losses.append(stop_flag)
             elif sig == utils.EXIT:
-                self._stop_losses[e] = take_flag
+                self.stop_losses[e] = take_flag
         utils.logger.debug('(%s) stop loss: %f pips, trader take profit: %f pips', self, stop_loss, take_profit)
 
     @utils.assert_logger
@@ -1074,7 +1078,7 @@ class Trader(object):
         """
         assert isinstance(credit_lev, (float, int)), 'credit_lev must be of type <float> or <int>'
 
-        self._credit_leverages = [credit_lev for i in range(len(self.df['Close']))]
+        self.credit_leverages = [credit_lev for i in range(len(self.df['Close']))]
         utils.logger.debug('(%s) credit leverage: %f', self, credit_lev)
 
     def get_support_resistance(self) -> Dict[str, Dict[int, float]]:
@@ -1104,17 +1108,42 @@ class Trader(object):
 
     def correct_sl_tp(self,
                       sl_correction: Union[float, int] = 50,
-                      tp_correction: Union[float, int] = 50):  # TODO: make documentation for this method   <-----------------------------------------------------------------------------------------------------------------
-        for e, (sl, tp, p, sig) in enumerate(zip(self._stop_losses, self._take_profits, self.df['Close'], self.returns)):
-            if sig == utils.BUY and sl > p:
-                self._stop_losses[e] = p * (1 - sl_correction / 10_000)
-            if sig == utils.SELL and sl < p:
-                self._stop_losses[e] = p * (1 + sl_correction / 10_000)
+                      tp_correction: Union[float, int] = 50):
+        stop_losses_before = self.stop_losses
+        take_profits_before = self.take_profits
+        for e, (sl, tp, p, sig, conv) in enumerate(zip(stop_losses_before, take_profits_before, self.df['Close'], self.returns, self._converted)):
+            if sig == utils.SELL:
+                if not np.isnan(conv):
+                    correct_sl = p * (1 + sl_correction / 10_000)
+                    correct_tp = p * (1 - tp_correction / 10_000)
+                    correct_sl_use = False
+                    correct_tp_use = False
 
-            if sig == utils.SELL and tp > p:
-                self._stop_losses[e] = p * (1 - tp_correction / 10_000)
-            if sig == utils.BUY and tp < p:
-                self._stop_losses[e] = p * (1 + tp_correction / 10_000)
+                if p < sl:
+                    correct_sl_use = True
+                if p > tp:
+                    correct_tp_use = True
+
+                if correct_sl_use:
+                    self.stop_losses[e] = sl
+                else:
+                    self.stop_losses[e] = correct_sl
+            elif sig == utils.BUY:
+                if not np.isnan(conv):
+                    correct_sl = p * (1 - sl_correction / 10_000)
+                    correct_tp = p * (1 + tp_correction / 10_000)
+                    correct_sl_use = False
+                    correct_tp_use = False
+
+                if p > sl:
+                    correct_sl_use = True
+                if p < tp:
+                    correct_tp_use = True
+
+                if correct_sl_use:
+                    self.stop_losses[e] = sl
+                else:
+                    self.stop_losses[e] = correct_sl
 
 
 class ExampleStrategies(Trader):
@@ -1126,6 +1155,7 @@ class ExampleStrategies(Trader):
                  **kwargs) -> List[Any]:
         return utils.get_window(self.df[column].values, n)
 
+    @strategy
     def find_pip_bar(self,
                      min_diff_coef: float = 2.0,
                      body_coef: float = 10.0) -> utils.PREDICT_TYPE_LIST:
@@ -1158,12 +1188,13 @@ class ExampleStrategies(Trader):
         self.set_open_stop_and_take()
         return self.returns
 
+    @strategy
     def find_DBLHC_DBHLC(self) -> utils.PREDICT_TYPE_LIST:
         self.returns = [utils.EXIT]
         flag: utils.PREDICT_TYPE = utils.EXIT
 
         flag_stop_loss: float = np.inf
-        self._stop_losses = [flag_stop_loss]
+        self.stop_losses = [flag_stop_loss]
         high: List[float]
         low: List[float]
         open_pr: List[float]
@@ -1183,11 +1214,12 @@ class ExampleStrategies(Trader):
                 flag_stop_loss = max(high[0], high[1])
 
             self.returns.append(flag)
-            self._stop_losses.append(flag_stop_loss)
+            self.stop_losses.append(flag_stop_loss)
         self.set_credit_leverages()
         self.set_open_stop_and_take(set_stop=False)
         return self.returns
 
+    @strategy
     def find_TBH_TBL(self) -> utils.PREDICT_TYPE_LIST:
         self.returns = [utils.EXIT]
         flag: utils.PREDICT_TYPE = utils.EXIT
@@ -1210,6 +1242,7 @@ class ExampleStrategies(Trader):
         self.set_open_stop_and_take()
         return self.returns
 
+    @strategy
     def find_PPR(self) -> utils.PREDICT_TYPE_LIST:
         self.returns = [utils.EXIT] * 2
         flag: utils.PREDICT_TYPE = utils.EXIT
@@ -1231,6 +1264,7 @@ class ExampleStrategies(Trader):
         self.set_open_stop_and_take()
         return self.returns
 
+    @strategy
     def strategy_ichimoku(self,
                           tenkansen: int = 9,
                           kijunsen: int = 26,
@@ -1290,7 +1324,7 @@ class ExampleStrategies(Trader):
                                name_slow=utils.SENKOU_SPAN_B_NAME)
 
             self.returns = [utils.EXIT for i in range(chinkouspan)]
-            self._stop_losses = [self.df['Close'].values[0]] * chinkouspan
+            self.stop_losses = [self.df['Close'].values[0]] * chinkouspan
             for e, (close, tenkan, kijun, A, B) in enumerate(zip(
                     prices.values[chinkouspan:],
                     tenkan_sen[chinkouspan:],
@@ -1326,11 +1360,11 @@ class ExampleStrategies(Trader):
                 min_cloud_now = min(senkou_span_a[e], senkou_span_b[e])
                 max_cloud_now = max(senkou_span_a[e], senkou_span_b[e])
                 if trade == utils.BUY:
-                    self._stop_losses.append(min_cloud_now)
+                    self.stop_losses.append(min_cloud_now)
                 elif trade == utils.SELL:
-                    self._stop_losses.append(max_cloud_now)
+                    self.stop_losses.append(max_cloud_now)
                 elif trade == utils.EXIT:
-                    self._stop_losses.append(0.0)
+                    self.stop_losses.append(0.0)
                 else:
                     raise ValueError('What???')
 
@@ -1339,12 +1373,14 @@ class ExampleStrategies(Trader):
         self.sl_tp_adder(add_stop_loss=stop_loss_plus)
         return self.returns
 
+    @strategy
     def strategy_buy_hold(self) -> utils.PREDICT_TYPE_LIST:
         self.returns = [utils.BUY for _ in range(len(self.df))]
         self.set_credit_leverages()
         self.set_open_stop_and_take()
         return self.returns
 
+    @strategy
     def strategy_2_sma(self,
                        slow: int = 100,
                        fast: int = 30,
@@ -1380,6 +1416,7 @@ class ExampleStrategies(Trader):
         self.set_credit_leverages()
         return self.returns
 
+    @strategy
     def strategy_3_sma(self,
                        slow: int = 100,
                        mid: int = 26,
@@ -1417,6 +1454,7 @@ class ExampleStrategies(Trader):
         self.set_open_stop_and_take()
         return self.returns
 
+    @strategy
     def strategy_3_ema(self,
                        slow: int = 46,
                        mid: int = 21,
@@ -1453,10 +1491,10 @@ class ExampleStrategies(Trader):
         self.set_open_stop_and_take()
         return self.returns
 
+    @strategy
     def strategy_macd(self,
                       slow: int = 100,
                       fast: int = 30) -> utils.PREDICT_TYPE_LIST:
-        self.returns = []
         diff = ta.trend.macd_diff(self.df['Close'], slow, fast)
 
         for j in diff:
@@ -1470,6 +1508,7 @@ class ExampleStrategies(Trader):
         self.set_open_stop_and_take()
         return self.returns
 
+    @strategy
     def strategy_rsi(self,
                      minimum: Union[float, int] = 13,
                      maximum: Union[float, int] = 87,
@@ -1495,13 +1534,14 @@ class ExampleStrategies(Trader):
         self.set_open_stop_and_take()
         return self.returns
 
+    @strategy
     def strategy_parabolic_SAR(self, plot: bool = True, **sar_kwargs) -> utils.PREDICT_TYPE_LIST:
         self.returns = []
         sar: ta.trend.PSARIndicator = ta.trend.PSARIndicator(self.df['High'], self.df['Low'],
                                                              self.df['Close'], **sar_kwargs)
         sardown: np.ndarray = sar.psar_down().values
         sarup: np.ndarray = sar.psar_up().values
-        self._stop_losses = sar.psar().values.tolist()
+        self.stop_losses = sar.psar().values.tolist()
         if plot:
             self.fig.plot_line(line=sarup,
                                width=utils.SAR_UP_WIDTH,
@@ -1534,6 +1574,7 @@ class ExampleStrategies(Trader):
         self.correct_sl_tp()
         return self.returns
 
+    @strategy
     def strategy_macd_histogram_diff(self,
                                      slow: int = 23,
                                      fast: int = 12,
@@ -1553,15 +1594,16 @@ class ExampleStrategies(Trader):
         self.set_open_stop_and_take()
         return self.returns
 
+    @strategy
     def strategy_supertrend(self,
                             plot: bool = True,
                             multiplier: float = 3.0,
                             length: int = 10) -> utils.PREDICT_TYPE_LIST:
-        st: utils.SuperTrendIndicator = utils.SuperTrendIndicator(self.df['Close'],
-                                                                  self.df['High'],
-                                                                  self.df['Low'],
-                                                                  multiplier=multiplier,
-                                                                  length=length)
+        st: indicators.SuperTrendIndicator = indicators.SuperTrendIndicator(self.df['Close'],
+                                                                            self.df['High'],
+                                                                            self.df['Low'],
+                                                                            multiplier=multiplier,
+                                                                            length=length)
         if plot:
             self.fig.plot_line(line=st.get_supertrend_upper(),
                                width=utils.ST_UP_WIDTH,
@@ -1580,12 +1622,13 @@ class ExampleStrategies(Trader):
                                _col=self.fig.data_col)
         self.returns = list(st.get_supertrend_strategy_returns())
         self.returns[:length + 1] = [utils.EXIT] * (length + 1)
-        self._stop_losses = list(st.get_supertrend())
-        self._stop_losses[0] = np.inf if self.returns[0] == utils.SELL else -np.inf
+        self.stop_losses = list(st.get_supertrend())
+        self.stop_losses[0] = np.inf if self.returns[0] == utils.SELL else -np.inf
         self.set_open_stop_and_take(set_stop=False)
         self.set_credit_leverages()
         return self.returns
 
+    @strategy
     def strategy_bollinger(self,
                            plot: bool = True,
                            to_mid: bool = False,
@@ -1639,10 +1682,11 @@ class ExampleStrategies(Trader):
             self.returns.append(flag)
         self.set_open_stop_and_take()
         if to_mid:
-            self._take_profits = mid_.tolist()
+            self.take_profits = mid_.tolist()
         self.set_credit_leverages()
         return self.returns
 
+    @strategy
     def strategy_bollinger_breakout(self,
                                     plot: bool = True,
                                     to_mid: bool = False,
@@ -1663,20 +1707,21 @@ class ExampleStrategies(Trader):
             mid_: pd.Series = bollinger.bollinger_mavg()
             upper: pd.Series = bollinger.bollinger_hband()
             lower: pd.Series = bollinger.bollinger_lband()
-            self._stop_losses = []
+            self.stop_losses = []
             for sig, high, low in zip(self.returns,
                                       upper,
                                       lower):
                 if sig == utils.BUY:
-                    self._stop_losses.append(low)
+                    self.stop_losses.append(low)
                 else:
-                    self._stop_losses.append(high)
+                    self.stop_losses.append(high)
         self.correct_sl_tp()
         return self.returns
 
+    @strategy
     def strategy_idris(self, points=20):
-        self._stop_losses = [np.inf] * 2
-        self._take_profits = [np.inf] * 2
+        self.stop_losses = [np.inf] * 2
+        self.take_profits = [np.inf] * 2
         flag = utils.EXIT
         self.returns = [flag] * 2
         for e in range(len(self.df) - 2):
@@ -1691,6 +1736,7 @@ class ExampleStrategies(Trader):
         self.set_credit_leverages()
         return self.returns
 
+    @strategy
     def DP_strategy(self,
                     length: int = 14,
                     s1: int = 3,
@@ -1715,6 +1761,7 @@ class ExampleStrategies(Trader):
                                     stop_loss=sl)
         return self.returns
 
+    @strategy
     def DP_2_strategy(self,
                       RSI_length: int = 14,
                       STOCH_length: int = 14,
@@ -1745,6 +1792,7 @@ class ExampleStrategies(Trader):
                                     stop_loss=sl)
         return self.returns
 
+    @strategy
     def strategy_kst(self, **kst_kwargs):
         KST = ta.trend.KSTIndicator(close=self.df['Close'], **kst_kwargs)
         fast = KST.kst()
@@ -1759,6 +1807,7 @@ class ExampleStrategies(Trader):
         self.set_open_stop_and_take(stop_loss=5000)
         return self.returns
 
+    @strategy
     def strategy_cci(self, **cci_kwargs):
         self.returns = []
         CCI = ta.trend.CCIIndicator(self.df['High'],
@@ -1772,8 +1821,9 @@ class ExampleStrategies(Trader):
         self.set_credit_leverages()
         self.set_open_stop_and_take()
 
+    @strategy
     def new_macd_strategy(self, slow=21, fast=12, ATR_win=14, ATR_multiplier=5):
-        self._stop_losses = []
+        self.stop_losses = []
         self.returns = []
 
         macd_indicator = ta.trend.MACD(close=self.df['Close'],
@@ -1795,10 +1845,10 @@ class ExampleStrategies(Trader):
 
             if diff > 0:
                 self.returns.append(utils.BUY)
-                self._stop_losses.append(price - stop_indicator)
+                self.stop_losses.append(price - stop_indicator)
             else:
                 self.returns.append(utils.SELL)
-                self._stop_losses.append(price + stop_indicator)
+                self.stop_losses.append(price + stop_indicator)
 
         self.set_open_stop_and_take(set_stop=False)
         self.set_credit_leverages()
