@@ -9,21 +9,21 @@ from typing import Sequence
 from typing import Sized
 from typing import Tuple
 from typing import Union
+from typing import Dict
 from warnings import warn
 
+import numpy as np
 import pandas as pd
 from numpy import arange
 from numpy import array
 from numpy import exp
 from numpy import isnan
-from numpy import log
 from numpy import nan
 from numpy import nan_to_num
 from numpy import ndarray
 from numpy import polyfit
-from pandas import DataFrame
 from pandas import Series
-from ta.volatility import AverageTrueRange
+from collections import defaultdict
 
 PREDICT_TYPE: type = int
 PREDICT_TYPE_LIST: type = List[PREDICT_TYPE]
@@ -162,6 +162,18 @@ LOWER_BB_COLOR: str = '#ff0000'
 LOWER_BB_WIDTH: float = 2
 LOWER_BB_ALPHA: float = 1.0
 
+STOP_BEFORE_TRAIN_NAME: str = 'train'
+STOP_BEFORE_TRAIN_COLOR: str = 'red'
+STOP_BEFORE_TRAIN_WIDTH: float = 1.0
+STOP_BEFORE_TRAIN_ALPHA: float = 1.0
+
+STOP_BEFORE_VAL_NAME: str = 'validation'
+STOP_BEFORE_VAL_COLOR: str = 'green'
+STOP_BEFORE_VAL_WIDTH: float = 1.0
+STOP_BEFORE_VAL_ALPHA: float = 1.0
+
+STOP_BEFORE_INTEGER_AS_INDEX: bool = True
+
 RESERVE: float = 0.99  # squeeze-protect
 
 INFO_TEXT: str = """losses: {}
@@ -176,14 +188,14 @@ calmar ratio: {}
 max drawdown: {}%
 profit/deviation ratio: {}"""  # .format(Trader.losses, Trader.trades, ...)
 
-__version__: str = "7.0.9"
+__version__: str = "7.7.8"
 __author__: str = 'Vlad Kochetov'
 __credits__: List[str] = [
     "Hemerson Tacon -- Stack overflow",
     "hpaulj -- Stack overflow",
     "furas -- Stack overflow",
     "Devin Jeanpierre (edit: wjandrea) -- Stack overflow",
-    "Войтенко Микола Полікарпович (Vojtenko Mykola Polikarpovich) -- helped me test the system of interaction with the binance crypto exchange",
+    "Voitenko Mykola Polikarpovich -- helped me test the system of interaction with the binance crypto exchange",
     "https://fxgears.com/index.php?threads/how-to-acquire-free-historical-tick-and-bar-data-for-algo-trading-and-backtesting-in-2020-stocks-forex-and-crypto-currency.1229/#post-19305 -- binance get historical data method",
     "https://www.geeksforgeeks.org/python-different-ways-to-kill-a-thread/ and https://teletype.in/@cozy_codespace/Hk70-Ntl4 -- heroku and threading problems",
     "https://stackoverflow.com/questions/57838939/handling-exceptions-with-bulk-api-requests -- IEX token",
@@ -193,7 +205,11 @@ __credits__: List[str] = [
     "Pavel Fedotov (https://github.com/Pfed-prog) -- pull request https://github.com/VladKochetov007/quick_trade/pull/60",
     "https://www.codearmo.com/blog/sharpe-sortino-and-calmar-ratios-python",
     "https://towardsdatascience.com/sharpe-ratio-sorino-ratio-and-calmar-ratio-252b0cddc328",
-    "https://www.wallstreetmojo.com/calmar-ratio/"
+    "https://www.wallstreetmojo.com/calmar-ratio/",
+    "Iliuschenko Nikita Andreevich -- translator",
+    "https://www.geeksforgeeks.org/append-to-json-file-using-python/",
+    "https://stackoverflow.com/questions/900392/getting-the-caller-function-name-inside-another-function-in-python",
+    "https://stackoverflow.com/questions/1263451/python-decorators-in-classes"
 ]
 
 logger = getLogger(__name__)
@@ -208,88 +224,25 @@ basicConfig(level=0,
                    'LINE %(lineno)d] %(name)s [%(processName)s: %(process)d] [%(threadName)s: %(thread)d] '
                    '[FILEPATH: %(pathname)s]\n')
 
+TUNER_INDENT: int = 2
+
+TUNER_CODECONF: Dict[str, str] = {
+    'winrate': 'winrate',
+    'trades': 'trades',
+    'losses': 'losses',
+    'profits': 'profits',
+    'percentage year profit': 'year_profit',
+    'mean deviation': 'mean_deviation',
+    'Sharpe ratio': 'sharpe_ratio',
+    'Sortino ratio': 'sortino_ratio',
+    'calmar ratio': 'calmar_ratio',
+    'max drawdown': 'max_drawdown',
+    'profit/deviation ratio': 'profit_deviation_ratio',
+}
+
 locker = threading.Lock()
 
 
-class SuperTrendIndicator(object):
-    """
-    Supertrend (ST)
-    """
-
-    close: Series
-    high: Series
-    low: Series
-
-    def __init__(self,
-                 close: Series,
-                 high: Series,
-                 low: Series,
-                 multiplier: float = 3.0,
-                 length: int = 10):
-        self.close = close
-        self.high = high
-        self.low = low
-        self.multiplier: float = multiplier
-        self.length = length
-        self._all = self._get_all_ST()
-
-    def get_supertrend(self) -> Series:
-        return self._all['ST']
-
-    def get_supertrend_upper(self) -> Series:
-        return self._all['ST_upper']
-
-    def get_supertrend_lower(self) -> Series:
-        return self._all['ST_lower']
-
-    def get_supertrend_strategy_returns(self) -> Series:
-        return self._all['ST_strategy']
-
-    def get_all_ST(self) -> DataFrame:
-        return self._all
-
-    def _get_all_ST(self) -> DataFrame:
-        m = self.close.size
-        dir_, trend = [1] * m, [0] * m
-        long, short = [nan] * m, [nan] * m
-        ATR = AverageTrueRange(high=self.high, low=self.low, close=self.close,
-                               window=self.length)
-
-        hl2_ = (self.high + self.low) / 2
-        matr = ATR.average_true_range() * self.multiplier
-        upperband = hl2_ + matr
-        lowerband = hl2_ - matr
-
-        for i in range(1, m):
-            if self.close.iloc[i] > upperband.iloc[i - 1]:
-                dir_[i] = BUY
-            elif self.close.iloc[i] < lowerband.iloc[i - 1]:
-                dir_[i] = SELL
-            else:
-                dir_[i] = dir_[i - 1]
-                if dir_[i] == BUY and lowerband.iloc[i] < lowerband.iloc[i - 1]:
-                    lowerband.iloc[i] = lowerband.iloc[i - 1]
-                if dir_[i] == SELL and upperband.iloc[i] > upperband.iloc[i - 1]:
-                    upperband.iloc[i] = upperband.iloc[i - 1]
-
-            if dir_[i] > 0:
-                trend[i] = long[i] = lowerband.iloc[i]
-            else:
-                trend[i] = short[i] = upperband.iloc[i]
-
-        df = DataFrame(
-            {
-                f"ST": trend,
-                f"ST_strategy": dir_,
-                f"ST_lower": long,
-                f"ST_upper": short,
-            },
-            index=self.close.index
-        )
-
-        return df
-
-
 def convert(data: PREDICT_TYPE_LIST) -> CONVERTED_TYPE_LIST:
     ret: List[Any] = list(data.copy())
     e: int
@@ -329,10 +282,18 @@ def convert_signal_str(predict: PREDICT_TYPE) -> str:
     elif predict == EXIT:
         return 'Exit'
 
+def log_list(values):
+    log_ret = []
+    for value in values:
+        if value >= 0:
+            log_ret.append(np.log(value))
+        else:
+            log_ret.append(-np.log(-value))
+    return array(log_ret)
 
 def get_exponential_growth(dataset: Sequence[float]) -> ndarray:
     x = arange(1, len(dataset) + 1, 1)
-    b, a = polyfit(x, log(array(dataset)), 1)
+    b, a = polyfit(x, log_list(array(dataset)), 1)
     regression = exp(a + b * x)
     return array(regression)
 
@@ -463,7 +424,8 @@ def get_diff(price: float,
 
 def make_multi_trade_returns(converted_returns: CONVERTED_TYPE_LIST) -> Tuple[PREDICT_TYPE_LIST, List[int]]:
     if EXIT in converted_returns:
-        warn('The use of utils.EXIT is deprecated in this type of strategy. If utils.EXIT is the first item in the sequence, you can replace it with np.nan.')
+        warn('The use of utils.EXIT is deprecated in this type of strategy. '
+             'If utils.EXIT is the first item in the sequence, you can replace it with np.nan.')
     result_credlev: List[int] = []
     result_returns: PREDICT_TYPE_LIST = [BUY] * len(converted_returns)
     flag_lev: int = 0
@@ -496,6 +458,27 @@ def get_multipliers(df: pd.Series) -> pd.Series:
     return ret
 
 
-def mean_deviation(frame: Series, avg_grwth: ndarray) -> float:
-    relative_diff = abs(frame.values - avg_grwth) / avg_grwth
+def mean_deviation(frame: Series, avg_growth: ndarray) -> float:
+    relative_diff = abs(frame.values - avg_growth) / avg_growth
     return relative_diff.mean()
+
+def year_profit(average_growth: np.ndarray, coef: Union[float, int]):
+    yp = profit_factor(average_growth) ** (coef - 1)
+    #  Compound interest. View https://www.investopedia.com/terms/c/compoundinterest.asp
+    yp -= 1  # The initial deposit does not count as profit
+    yp *= 100  # Percentage
+    return yp
+
+def map_dict(func, data):
+    if isinstance(data, dict):
+        for key, val in data.items():
+            if isinstance(val, dict):
+                data[key] = func(val)
+                data[key] = map_dict(func, data[key])
+    return func(data)
+
+
+def recursive_dict(base={}):
+    get_dd = lambda: defaultdict(get_dd)
+    get_dd_base = lambda b: defaultdict(get_dd, b)
+    return map_dict(get_dd_base, base)
